/**************************************************************************
* /dev/iptrk device for Linux 2.4 kernel
* Copyright (C) 2003 Jason Nunn <jsno@leal.com.au>
* Darwin, NT, Australia. 
* Originally developed for Arafura Connect (Dec 2000 - Mar 2003)
* Donated to GPL before liquidation.
*
* =======================================================================
* /dev/iptrk interface
*
**************************************************************************/
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/vmalloc.h>
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>
#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
#endif
#include <linux/iptrk.h>
#include "../../version.h"

#define IPTRK_MAJOR 201

typedef struct
{
  unsigned mem_f   : 1;
  unsigned devfs_f : 1;
  unsigned dev_f   : 1;
} iptrk_flags;

static char           *device_label = "iptrk";
static int            iptrk_queue_optr,
                      iptrk_queue_size_bytes;
static atomic_t       open_s_count;
static spinlock_t     open_s_lock;
static iptrk_flags    flags;

#ifdef CONFIG_DEVFS_FS
static devfs_handle_t devfs_h;
#endif

int qsize = 30000;
MODULE_PARM(qsize, "i");
MODULE_PARM_DESC(qsize, "iptrk circular queue size");
MODULE_AUTHOR("Jason Nunn");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

/**************************************************************************
* /proc code
**************************************************************************/
static int read_procmem(
  char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
  int           len = 0;
  unsigned long ifg;

  if(atomic_read(&iptrk_active_f))
  {
    spin_lock_irqsave(&iptrk_lock, ifg);  /*sprintf() doesn't sleep.. 
                                            ok to use in spin_lock*/
    len += sprintf(buf,
      "queue_info:\n"
      "\tsize:\t%d (%d bytes)\n"
      "\tutil:\t%d (%d%%)\n"
      "\tOvfw:\t%d\n",
      iptrk_queue_size,
      iptrk_queue_size_bytes,
      iptrk_queue_count,
      (iptrk_queue_count * 100) / iptrk_queue_size, 
      iptrk_overflow_count);

    spin_unlock_irqrestore(&iptrk_lock, ifg);
  }
  else
  {
    len += sprintf(buf, "off\n");
  }

  *eof = 1;
  return len;
}

/**************************************************************************
* io code
**************************************************************************/
static ssize_t read(
  struct file *file, char *buf, size_t length, loff_t *ppos)
{
  unsigned long ifg, i;

  spin_lock_irqsave(&iptrk_lock, ifg);

  if(iptrk_queue_count == 0)  /*no need to test iptrk_active_f*/
  {
    spin_unlock_irqrestore(&iptrk_lock, ifg);
    return 0;
  }

  i = copy_to_user(buf, &iptrk_queue[iptrk_queue_optr], sizeof(iptrk_queue_t));
  iptrk_queue_optr++;
  iptrk_queue_optr %= iptrk_queue_size;
  iptrk_queue_count--;

  spin_unlock_irqrestore(&iptrk_lock, ifg);

  if(i)
    return -EFAULT;
  return sizeof(iptrk_queue_t);
}

static ssize_t write(
  struct file *file, const char *buf, size_t count, loff_t *ppos)
{
  if(*buf == '1')
  {
    atomic_set(&iptrk_active_f, 1);
    printk(KERN_NOTICE "iptrk: active.\n");
  }
  else
  {
    atomic_set(&iptrk_active_f, 0);
    printk(KERN_NOTICE "iptrk: off.\n");
  }
  return count;
}

/*
  single access code based on p165, Linux device drivers 2nd ed,
  Rubini et al
*/
static int open_s(struct inode *ip, struct file *fp)
{
  if(MINOR(ip->i_rdev) > 0)
    return -ENODEV;

  spin_lock(&open_s_lock);
  if(atomic_read(&open_s_count))
  {
    spin_unlock(&open_s_lock);
    return -EBUSY;
  }
  atomic_set(&open_s_count, 1);
  spin_unlock(&open_s_lock);
  MOD_INC_USE_COUNT;

  return 0;
}

/*
  ditto, accept i'm using an atomic flag here
*/
static int release(struct inode *ip, struct file *fp)
{
  atomic_set(&open_s_count, 0); /*don't need to spin_lock, as only one 
                                  instance is going to decrement it*/
  MOD_DEC_USE_COUNT;
  return 0;
}

static int ioctl(
  struct inode *ip, struct file *fp, unsigned int cmd, unsigned long arg)
{
  return -EINVAL;
}

/**************************************************************************
* init code
**************************************************************************/
static struct file_operations fops =
{
  read:    read,
  write:   write,
  open:    open_s,
  release: release,
  ioctl:   ioctl
};

static void release_resources(void)
{
  unsigned long ifg;

 /*
  * block until nobody else is using queue. then deactivate. then unlock.
  * once inactive, nobdy can use it. now free to deallocate queue.
  */
  spin_lock_irqsave(&iptrk_lock, ifg);
  atomic_set(&iptrk_active_f, 0);
  spin_unlock_irqrestore(&iptrk_lock, ifg);

  remove_proc_entry(device_label, NULL);

#ifdef CONFIG_DEVFS_FS
  if(flags.devfs_f)
    devfs_unregister(devfs_h);
#endif

  if(flags.dev_f)
    unregister_chrdev(IPTRK_MAJOR, device_label);

  if(flags.mem_f)
    vfree(iptrk_queue);
}

static int iptrk_init(void)
{
  SET_MODULE_OWNER(&fops);

  printk(KERN_NOTICE "iptrk: " WELCOME);
  printk(KERN_NOTICE "iptrk: iptrk kernel rev: " IPTRK_KERNEL_REVISION "\n");

  if(qsize < 50)
  {
    printk(
      KERN_NOTICE "iptrk: queue buffer size too small (%d entries), aborting\n",
      iptrk_queue_size);
    return 0;
  }

  spin_lock_init(&open_s_lock);
  iptrk_queue_size = qsize;
  iptrk_queue_size_bytes = iptrk_queue_size * sizeof(iptrk_queue_t);
  iptrk_queue = vmalloc(iptrk_queue_size_bytes);
  if(!iptrk_queue)
  {
    printk(KERN_ERR "iptrk: error: allocating queue buffer\n");
    release_resources();
    return -ENOMEM;
  }
  flags.mem_f = 1;

#ifdef CONFIG_DEVFS_FS
  devfs_h = devfs_register(
    NULL,
    device_label, DEVFS_FL_DEFAULT,
    IPTRK_MAJOR, 0, S_IFCHR | S_IRUSR | S_IWUSR, &fops, NULL);
  if(!devfs_h)
  {
    printk("iptrk: unable to register driver (devfs_register)\n");
    release_resources();
    return -EBUSY;
  }
  flags.devfs_f = 1;
#endif
  if(register_chrdev(IPTRK_MAJOR, device_label, &fops) < 0)
  {
    printk("iptrk: unable to register driver (register_chrdev)\n");
    release_resources();
    return -EIO;
  }
  flags.dev_f = 1;

  iptrk_queue_count = 0;
  iptrk_overflow_count = 0;
  iptrk_queue_iptr = 0;
  iptrk_queue_optr = 0;

  create_proc_read_entry(device_label, 0, NULL, read_procmem, NULL);
  return 0;
}

static void iptrk_cleanup(void)
{
  release_resources();
  printk(KERN_NOTICE "iptrk: uninstalled.\n");
}

module_init(iptrk_init);
module_exit(iptrk_cleanup);
